/* -----------------------------------------------------------------------------
 * Copyright 2022 Massachusetts Institute of Technology.
 * All Rights Reserved
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  1. Redistributions of source code must retain the above copyright notice,
 *     this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright notice,
 *     this list of conditions and the following disclaimer in the documentation
 *     and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Research was sponsored by the United States Air Force Research Laboratory and
 * the United States Air Force Artificial Intelligence Accelerator and was
 * accomplished under Cooperative Agreement Number FA8750-19-2-1000. The views
 * and conclusions contained in this document are those of the authors and should
 * not be interpreted as representing the official policies, either expressed or
 * implied, of the United States Air Force or the U.S. Government. The U.S.
 * Government is authorized to reproduce and distribute reprints for Government
 * purposes notwithstanding any copyright notation herein.
 * -------------------------------------------------------------------------- */
#include "spark_dsg/layer_view.h"

namespace spark_dsg {

using ViewIter = LayerView::Iter;

ViewIter::Iter(const SceneGraphLayer* layer, const Filter& filter, bool is_end)
    : layer_(layer),
      iter_(is_end ? layer->nodes().end() : layer->nodes().begin()),
      filter_(filter) {
  if (!valid()) {
    next();  // noop if at end already
  }
}

ViewIter::Iter(const Iter& other)
    : layer_(other.layer_), iter_(other.iter_), filter_(other.filter_) {}

ViewIter& ViewIter::operator=(const ViewIter& other) {
  layer_ = other.layer_;
  iter_ = other.iter_;
  filter_ = other.filter_;
  return *this;
}

bool ViewIter::valid() const {
  if (iter_ == layer_->nodes().end()) {
    return false;
  }

  if (!filter_) {
    return true;
  }

  return filter_(*iter_->second);
}

void ViewIter::next() {
  if (iter_ == layer_->nodes().end()) {
    return;
  }

  // increment to next available active node
  do {
    ++iter_;
    if (iter_ == layer_->nodes().end()) {
      break;
    }
  } while (!valid());
}

ViewIter& ViewIter::operator++() {
  next();
  return *this;
}

ViewIter ViewIter::operator++(int) {
  auto tmp = *this;
  ++(*this);
  return tmp;
}

ViewIter::reference ViewIter::operator*() const { return *iter_->second; }

ViewIter::pointer ViewIter::operator->() const { return iter_->second.get(); }

LayerView::LayerView() : layer_(nullptr) {}

LayerView::LayerView(const SceneGraphLayer& layer, const Filter& filter)
    : layer_(&layer), filter_(filter) {}

ViewIter LayerView::begin() const { return Iter(layer_, filter_, false); }

ViewIter LayerView::end() const { return Iter(layer_, filter_, true); }

bool operator==(const ViewIter& lhs, const ViewIter& rhs) {
  return lhs.layer_ == rhs.layer_ && lhs.iter_ == rhs.iter_;
}

}  // namespace spark_dsg
